/* 
 * Copyright (C) 2020 wea_ondara
 *
 * BungeePerms is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BungeePerms is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.alpenblock.bungeeperms.config;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

public class ConfigurationSerialization
{

    public static final String SERIALIZED_TYPE_KEY = "==";
    private final Class<? extends ConfigurationSerializable> clazz;
    private static Map<String, Class<? extends ConfigurationSerializable>> aliases = new HashMap<String, Class<? extends ConfigurationSerializable>>();

    protected ConfigurationSerialization(Class<? extends ConfigurationSerializable> clazz)
    {
        this.clazz = clazz;
    }

    protected Method getMethod(String name, boolean isStatic)
    {
        try
        {
            Method method = clazz.getDeclaredMethod(name, Map.class);

            if (!ConfigurationSerializable.class.isAssignableFrom(method.getReturnType()))
            {
                return null;
            }
            if (Modifier.isStatic(method.getModifiers()) != isStatic)
            {
                return null;
            }

            return method;
        }
        catch (NoSuchMethodException ex)
        {
            return null;
        }
        catch (SecurityException ex)
        {
            return null;
        }
    }

    protected Constructor<? extends ConfigurationSerializable> getConstructor()
    {
        try
        {
            return clazz.getConstructor(Map.class);
        }
        catch (NoSuchMethodException ex)
        {
            return null;
        }
        catch (SecurityException ex)
        {
            return null;
        }
    }

    protected ConfigurationSerializable deserializeViaMethod(Method method, Map<String, Object> args)
    {
        try
        {
            ConfigurationSerializable result = (ConfigurationSerializable) method.invoke(null, args);

            if (result == null)
            {
                Logger.getLogger(ConfigurationSerialization.class.getName()).log(Level.SEVERE, "Could not call method '" + method.toString() + "' of " + clazz + " for deserialization: method returned null");
            }
            else
            {
                return result;
            }
        }
        catch (Throwable ex)
        {
            Logger.getLogger(ConfigurationSerialization.class.getName()).log(
                    Level.SEVERE,
                    "Could not call method '" + method.toString() + "' of " + clazz + " for deserialization",
                    ex instanceof InvocationTargetException ? ex.getCause() : ex);
        }

        return null;
    }

    protected ConfigurationSerializable deserializeViaCtor(Constructor<? extends ConfigurationSerializable> ctor, Map<String, Object> args)
    {
        try
        {
            return ctor.newInstance(args);
        }
        catch (Throwable ex)
        {
            Logger.getLogger(ConfigurationSerialization.class.getName()).log(
                    Level.SEVERE,
                    "Could not call constructor '" + ctor.toString() + "' of " + clazz + " for deserialization",
                    ex instanceof InvocationTargetException ? ex.getCause() : ex);
        }

        return null;
    }

    public ConfigurationSerializable deserialize(Map<String, Object> args)
    {

        ConfigurationSerializable result = null;
        Method method = null;

        if (result == null)
        {
            method = getMethod("deserialize", true);

            if (method != null)
            {
                result = deserializeViaMethod(method, args);
            }
        }

        if (result == null)
        {
            method = getMethod("valueOf", true);

            if (method != null)
            {
                result = deserializeViaMethod(method, args);
            }
        }

        if (result == null)
        {
            Constructor<? extends ConfigurationSerializable> constructor = getConstructor();

            if (constructor != null)
            {
                result = deserializeViaCtor(constructor, args);
            }
        }

        return result;
    }

    public static ConfigurationSerializable deserializeObject(Map<String, Object> args, Class<? extends ConfigurationSerializable> clazz)
    {
        return new ConfigurationSerialization(clazz).deserialize(args);
    }

    public static ConfigurationSerializable deserializeObject(Map<String, Object> args)
    {
        Class<? extends ConfigurationSerializable> clazz = null;

        if (args.containsKey(SERIALIZED_TYPE_KEY))
        {
            try
            {
                String alias = (String) args.get(SERIALIZED_TYPE_KEY);

                if (alias == null)
                {
                    throw new IllegalArgumentException("Cannot have null alias");
                }
                clazz = getClassByAlias(alias);
                if (clazz == null)
                {
                    throw new IllegalArgumentException("Specified class does not exist ('" + alias + "')");
                }
            }
            catch (ClassCastException ex)
            {
                ex.fillInStackTrace();
                throw ex;
            }
        }
        else
        {
            throw new IllegalArgumentException("Args doesn't contain type key ('" + SERIALIZED_TYPE_KEY + "')");
        }

        return new ConfigurationSerialization(clazz).deserialize(args);
    }

//    public static void registerClass(Class<? extends ConfigurationSerializable> clazz) {
//        DelegateDeserialization delegate = clazz.getAnnotation(DelegateDeserialization.class);
//
//        if (delegate == null) {
//            registerClass(clazz, getAlias(clazz));
//            registerClass(clazz, clazz.getName());
//        }
//    }
    public static void registerClass(Class<? extends ConfigurationSerializable> clazz, String alias)
    {
        aliases.put(alias, clazz);
    }

    public static void unregisterClass(String alias)
    {
        aliases.remove(alias);
    }

    public static void unregisterClass(Class<? extends ConfigurationSerializable> clazz)
    {
        while (aliases.values().remove(clazz))
        {
            ;
        }
    }

    public static Class<? extends ConfigurationSerializable> getClassByAlias(String alias)
    {
        return aliases.get(alias);
    }

    public static String getAlias(Class<? extends ConfigurationSerializable> clazz)
    {
        DelegateDeserialization delegate = clazz.getAnnotation(DelegateDeserialization.class);

        if (delegate != null)
        {
            if ((delegate.value() == null) || (delegate.value() == clazz))
            {
                delegate = null;
            }
            else
            {
                return getAlias(delegate.value());
            }
        }

        if (delegate == null)
        {
            SerializableAs alias = clazz.getAnnotation(SerializableAs.class);

            if ((alias != null) && (alias.value() != null))
            {
                return alias.value();
            }
        }

        return clazz.getName();
    }
}
